require "numru/gphys"
require "no_rdb_base"

class AnalysisColumn < ActiveRecord::ConnectionAdapters::Column

  def type_cast_code(var_name)
    case type
    when :array
      "#{self.class.name}.value_to_array(#{var_name})"
    when :array_int
      "#{self.class.name}.value_to_array(#{var_name}, :int)"
    when :array_float
      "#{self.class.name}.value_to_array(#{var_name}, :float)"
    when :model
      "#{self.class.name}.check_model(#{var_name})"
    else
      super
    end
  end

  def self.value_to_array(value, val_type=nil)
    if Array === value
      ary = value
    elsif Hash === value
      ary = Array.new
      value.each{|k,v|
        ary[k.to_i] = v
      }
    elsif String === value
      ary = value.split(/,/)
    else
      ary = [value]
    end
    case val_type
    when :int
      ary.collect!{|v| v.to_i}
    when :float
      ary.collect!{|v| v.to_f}
    end
    return ary
  end

  def self.check_model(value)
    if ActiveRecord::Base === value
      return value
    else
      raise "a model is expected"
    end
  end
    

  private
  def simplified_type(type)
    case type
    when /array_int/i
      :array_int
    when /array_float/i
      :array_float
    when /array/i
      :array
    when /model/i
      :model
    else
      super(type)
    end
  end

end

class Analysis < ActiveRecord::NoRdbBase

  ACTION_TYPE = %w( draw analysis )

  DRAW_TYPE = %w( line tone vector )

  DRAW_PROJECTION = {1 => "rectangular uniform coordinate",
                     2 => "semi-logarithmic coordinate (y axis)",
                     3 => "semi-logarithmic coordinate (x axis)",
                     4 => "logarithmic coordinate",
                     5 => "polar coordinate",
                     6 => "bipolar coordinate",
#                     7 => "elliptic coordinate",
                     10 => "equidistant cylindrical projection",
                     11 => "Mercator's projection",
                     12 => "Mollweide's projection",
                     13 => "Hammer's projection",
                     14 => "Eckert VI projection",
                     15 => "Kitada's elliptic projection",
                     20 => "equidistant conical projection",
                     21 => "Lambert's equal-area conical projection",
                     22 => "Lambert's conformal conical projection",
                     23 => "Bonne's projection",
                     30 => "orthographic projection",
                     31 => "polar stereo projection",
                     32 => "azimuthal equidistant projection",
                     33 => "Lambert's azimuthal equal-area projection"
                    }

  DRAW_SIZE = [[700,700], [400,400], [250,250]]

  common_attrs = [
     ["variables", [], "array", false],
     ["action_type", ACTION_TYPE.index("draw"), "int", false]
  ]
  draw_common_attrs = [
    ["draw_share", true, "boolean", false],
    ["draw_type", DRAW_TYPE.index("tone"), "int", false],
    ["x_axis", 0, "int", false],
    ["y_axis", 1, "int"],
    ["draw_projection", 0, "int", false],
    ["region", [], "array", false],
    ["draw_keep", false, "boolean", false],
    ["draw_size", 1, "int", false]
  ]
  @@draw_line_attrs = [
  ]
  @@draw_tone_attrs = [
    ["tone", true, "boolean", false, :vizshot],
    ["contour", true, "boolean", false, :vizshot],
    ["min", nil, "float", true, :ggraph],
    ["max", nil, "float", true, :ggraph],
    ["nlev", nil, "int", true, :ggraph],
    ["levels", nil, "array_float", true, :ggraph],
    ["color_bar", true, "boolean", false, :vizshot]
  ]
  @@draw_vector_attrs = [
    ["vector_variables_order", [], "array", false]
  ]
  function_attrs = [
    ["function", nil, "model", false],
    ["input_num", nil, "int", false],
    ["function_arguments", [], "array", false],
    ["function_variables_order", [], "array", false]
  ]

  @@minmaxs = {
    "action_type" => [0, ACTION_TYPE.length-1],
    "draw_type" => [0, DRAW_TYPE.length-1],
    "draw_size" => [0, DRAW_SIZE.length-1],
    "input_num" => [1, Float::MAX]
  }



  (common_attrs +
   draw_common_attrs +
   @@draw_line_attrs + @@draw_tone_attrs + @@draw_vector_attrs +
   function_attrs
  ).each{|ary|
    push_column AnalysisColumn.new( *ary[0..3] )
  }

  def write_attribute(attr_name, value)
    case attr_name.to_s
    when "function"
      unless Function === value
        begin
          value = Function.find(value)
        rescue ActiveRecord::RecordNotFound
          value = nil
        end
      end
    when "vector_variables_order"
      value = variables_order(value)
    when "function_variables_order"
      value = variables_order(value)
    end

    super

    if (minmax = @@minmaxs[attr_name.to_s])
      val = read_attribute(attr_name)
      if val < minmax[0]
        write_attribute(attr_name, minmax[0].to_s)
      elsif val > minmax[1]
        write_attribute(attr_name, minmax[1].to_s)
      end
    end
    case attr_name.to_s
    when "x_axis"
      if (xa = read_attribute("x_axis")) == read_attribute("y_axis")
        write_attribute("y_axis", xa==0 ? "1" : "0")
      end
    when "y_axis"
      if (ya = read_attribute("y_axis")) == read_attribute("x_axis")
        write_attribute("x_axis",ya==0 ? "1" : "0")
      end
    when "draw_projection"
      unless DRAW_PROJECTION[read_attribute("draw_projection")]
        write_attribute("draw_projection", "1")
      end
    when "levels"
      str = read_attribute("levels")
      unless (str_new=str.split(/,/).collect{|x| x.to_f}.join(',')) == str
        write_attribute("levels", str_new)
                     # to ensure comma separated numerics
      end
    end
  end

  def variable_clear
    @dimensions = nil
    %w( variables region x_axis y_axis vector_variables_order function_variables_order ).each{|name|
      column = column_for_attribute(name)
      self[name] = column.default
    }
  end

  def file_and_variable_names
    variables.collect do |var|
      var.file_and_variable_names
    end
  end

  def gphyses
    variables.collect do |var|
      var.gphys
    end
  end

  def dimensions
    unless @dimensions
      gphys = self.gphyses
      if gphys.length == 0
        @dimensions = []
      else
        gp = gphys[0]
        axes = Array.new
        gp.rank.times do |i|
          axes.push gp.axis(i)
        end
        flag = true
        (gphys.length-1).times do |i|
          gp = gphys[i+1]
          unless axes.length == gp.rank
            flag = false
            break
          end
          axes.length.times do |j|
            axis = axes[j]
            axis0 = gp.axis(j)
            unless axis.name == axis0.name
              flag = false
              break
            end
#            unless axis.pos.val == axis0.pos.val
#              flag = false
#              break
#            end
          end
          unless flag
            break
          end
        end
        if flag
          @dimensions = axes.collect do |axis|
            pos = axis.pos
            {:name => axis.name, :units => pos.units, :ary => pos.val.to_a}
          end
        else
          @dimensions = []
        end
      end
    end
    return @dimensions
  end

  def index
    ary = Array.new
    dims = region
    dims.length.times do |i|
      dim = dims[i]
      if ACTION_TYPE[action_type]=="analysis" || (i==x_axis || (DRAW_TYPE[draw_type]!="line"&&i==y_axis))
        ary.push dim[:min].to_f..dim[:max].to_f
      else
        ary.push dim[:min].to_f
      end
    end
    return ary
  end

  def diagram_label
    return nil unless ACTION_TYPE[action_type]=="draw"
    return nil if variables.nil? || variables.length==0
    label = DRAW_TYPE[draw_type]
    label += ", #{dimensions[x_axis][:name]}(#{region[x_axis][:min]}:#{region[x_axis][:max]})"
    unless DRAW_TYPE[draw_type] == "line"
    label += " vs #{dimensions[y_axis][:name]}(#{region[y_axis][:min]}:#{region[y_axis][:max]})"
    end
    tmp = Array.new
    dimensions.length.times{|i|
      next if i==x_axis
      next if DRAW_TYPE[draw_type]!="line" && i==y_axis
      tmp.push "#{dimensions[i][:name]}=#{region[i][:min]}"
    }
    label += " @ "+tmp.join(",") unless tmp.length==0
    return label
  end


  def get_vizshots(work_prefix)
    options = Hash.new
    vars = variables
    nvars = variables.length
    case DRAW_TYPE[draw_type]
    when "line"
      if nvars == 0
        return [false, "at least one variable must be selected"]
      end
      options[:method] = :line

      options.update get_vizshot_options_from_attrs(@@draw_line_attrs)

      nfig = nvars
    when "tone"
      if nvars == 0
        return [false, "at least one variable must be selected"]
      elsif ! (tone || contour)
        return [false, "neither tone or contour are not selected"]
      end
      options[:method] = :tone_cont

      if x_axis > y_axis
        options["transpose"] = true
      end

      options.update get_vizshot_options_from_attrs(@@draw_tone_attrs)

      nfig = nvars
    when "vector"
      if nvars%2 != 0
        return [false, "number of variable must be multiples of 2"]
      end
      options[:method] = :vector
      options["transpose"] = true if x_axis > y_axis

      options.update get_vizshot_options_from_attrs(@@draw_vector_attrs)

      nfig = nvars/2
    else
      raise "[BUG] figure type is invalid"
    end
    options[:cut] = index
    size = Analysis::DRAW_SIZE[draw_size]
    vizs = Array.new
    nfig.times{|n|
      if Analysis::DRAW_TYPE[draw_type]=="vector"
        options[:variable] = vars[vector_variables_order[n][0]]
        options[:variable2] = vars[vector_variables_order[n][1]]
      else
        options[:variable] = vars[n]
      end
      viz = NumRu::VizShot.new(:iwidth => size[0], :iheight => size[1], :basename => File.basename(work_prefix))
      viz.set_fig('itr' => draw_projection)
      viz.set_tone('tonf' => true )
      viz.plot( options.dup )
      vizs.push viz
    }
    return [true, vizs]
  end



  def self.from_vizshot(viz)
    plot = viz.get_plots
    return nil unless plot.length==1
    plot = plot.first
    plot.each{|k,v|
      if String === k
        plot[k.to_sym] = v
      end
    }

    vars = viz.get_variables

    analysis Analysis.new

    vars.each{|var|
      return nil unless var.id
      analysis.variables.push var
    }
    analysis.action_type = ACTION_TYPE.index("draw")
    axis = Array.new
    plot[:cut].each_with_index{|range,i|
      if Range === range
        analysis.region( i => {"min"=>range.first,"max"=>range.last} )
        axis.push i
      elsif Numeric === range
        analysis.region( i => {"min"=>range} )
      else
        raise "[BUG]"
      end
    }
    case plot[:method]
    when :line
      type = 0
      unless axis.length==1
        raise "[BUG]"
      end
      twoD = false
    when :tone_cont
      type = 1
      unless axis.length==2
        raise "[BUG]"
      end
      twoD = true
    when :vector
      type = 2
      unless axis.length==2
        raise "[BUG]"
      end
      twoD = true
    else
      raise "[BUG]"
    end
    analysis.draw_type = type
    axis = [axis[1],axis[0]] if plot[:transpose] if twoD
    analysis.x_axis = axis[0]
    analysis.y_axis = axis[1] if twoD
    analysis.draw_projection = viz.get_itr
    isize = DRAW_SIZE.index(viz.get_size)
    unless isize
      raise "[BUG]"
    end
    analysis.draw_size = isize
    case plot[:method]
    when :line
      analysis.set_attrs_from_vizshot_plot(@@draw_line_attrs, plot)
    when :tone_cont
      analysis.set_attrs_from_vizshot_plot(@@draw_tone_attrs, plot)
    when :vector
      analysis.set_attrs_from_vizshot_plot(@@draw_vector_attrs, plot)
      analysis.vector_variables_order = {0 => {0=>0,1=>1} }
    end
    analysis
  end


  def self.uri_params_from_vizshot(viz)
    plot = viz.get_plots
    return nil unless plot.length==1
    plot = plot.first
    plot.each{|k,v|
      if String === k
        plot[k.to_sym] = v
      end
    }

    vars = viz.get_variables

    params = Array.new
    vars.each{|var|
      return nil unless var.id
      params.push "variables[#{var.path}]=1"
    }
    params.push "action_type=draw"
    axis = Array.new
    plot[:cut].each_with_index{|range,i|
      if Range === range
        params.push "analysis[region][#{i}][min]=#{range.first}"
        params.push "analysis[region][#{i}][max]=#{range.last}"
        axis.push i
      elsif Numeric === range
        params.push "analysis[region][#{i}][min]=#{range}"
      else
        raise "[BUG]"
      end
    }
    case plot[:method]
    when :line
      type = 0
      unless axis.length==1
        raise "[BUG]"
      end
      twoD = false
    when :tone_cont
      type = 1
      unless axis.length==2
        raise "[BUG]"
      end
      twoD = true
    when :vector
      type = 2
      unless axis.length==2
        raise "[BUG]"
      end
      twoD = true
    else
      raise "[BUG]"
    end
    params.push "analysis[draw_type]=#{type}"
    axis = [axis[1],axis[0]] if plot[:transpose] if twoD
    params.push "analysis[x_axis]=#{axis[0]}"
    params.push "analysis[y_axis]=#{axis[1]}" if twoD
    params.push "analysis[draw_projection]=#{viz.get_itr}"
    isize = DRAW_SIZE.index(viz.get_size)
    unless isize
      raise "[BUG]"
    end
    params.push "analysis[draw_size]=#{isize}"
    case plot[:method]
    when :line
      params += get_uri_params_from_vizshot_plot(@@draw_line_attrs, plot)
    when :tone_cont
      params += get_uri_params_from_vizshot_plot(@@draw_tone_attrs, plot)
    when :vector
      params += get_uri_params_from_vizshot_plot(@@draw_vector_attrs, plot)
      params.push "analysis[vector_variables_order][0][0]=0"
      params.push "analysis[vector_variables_order][0][1]=1"
    end
#    params.collect{|str| str.gsub(/[/,"%5B").gsub(/]/,"%5D")}.join("&")
    URI.escape( params.join("&") )
  end


  def self.get_uri_params_from_vizshot_plot(attrs, plot)
    ary = Array.new
    attrs.each{|attr|
      case attr[4]
      when :vizshot
        name = attr[0].to_sym
      when :ggraph
        name = attr[0]
      else
        next
      end
      val = plot[name]
      case attr[2]
      when "boolean"
        val = val ? 1 : 0
      when /array/
        val &&= val.join(",")
      end
      if !attr[3] || !val.nil?
        ary.push "analysis[#{attr[0]}]=#{val.inspect}"
      end
    }
    return ary
  end

  private
  def variables_order(val)
    unless Hash === val
      return nil
    end
    variables_order = Hash.new
    val.each{|k,v|
      if Hash === v
        tmp = Array.new
        v.each{|k1,v1|
          tmp[k1.to_i] = v1.to_i
        }
        variables_order[k.to_i] = tmp
      else
        return nil
      end
    }
    ary = variables_order.values.flatten
    unless ary.max == ary.length-1
      return nil
    end
    ary.length.times{|i|
      unless ary.include?(i)
        return nil
      end
    }
    return variables_order
  end


  def get_vizshot_options_from_attrs(attrs)
    hash = Hash.new
    attrs.each{|attr|
      case attr[4]
      when :vizshot
        name = attr[0].to_sym
      when :ggraph
        name = attr[0]
      else
        next
      end
      val = self.send(attr[0])
      if !attr[3] || (!val.nil? && val!=[] && val!="")
        hash[name] = val
      end
    }
    return hash
  end

  def set_attrs_from_vizshot_plot(attrs, plot)
    attrs.each{|attr|
      case attr[4]
      when :vizshot
        name = attr[0].to_sym
      when :ggraph
        name = attr[0]
      else
        next
      end
      val = plot[name]
      unless val.nil?
        self.send(attr[0]+"=", val)
      end
    }
  end

end
